@@ -1,12 +1,18 @@
 """Utilities for extracting keyword information from search engine
 referrers."""
+from __future__ import absolute_import, division, print_function, unicode_literals
+
+import logging
 import re
-import logging
-from itertools import groupby
-from urlparse import urlparse, parse_qs, ParseResult
+import sys
+from collections import defaultdict
+from io import TextIOWrapper
 
-from iso3166 import countries
 import pylru
+import tldextract
+from iso3166 import countries
+from six import iteritems, itervalues, PY3, string_types, text_type
+from six.moves.urllib.parse import urlparse, parse_qs, ParseResult
 
 # import pkg_resources
 # with fallback for environments that lack it
@@ -25,16 +31,16 @@ except ImportError:
             f = os.path.join(moddir, resource_name)
             return open(f)
 
-# import cPickle
-# for performance with a fallback on Python pickle
+# import ujson for performance with a fallback on default json
 try:
-    import cPickle as pickle
+    import ujson as json
 except ImportError:
-    import pickle
+    import json
 
 
 __all__ = ('get_parser', 'is_serp', 'extract', 'get_all_query_params',
-           'add_custom_parser', 'SearchEngineParser')
+           'get_all_query_params_by_domain', 'add_custom_parser',
+           'SearchEngineParser')
 
 log = logging.getLogger('serpextract')
 
@@ -45,6 +51,12 @@ _country_codes = [country.alpha2.lower()
 # http://en.wikipedia.org/wiki/ISO_3166-1 for more information
 _country_codes += ['uk']
 
+# For generating possible variations of domains based
+_second_level_domains = ['co', 'com']
+
+# Cache for querystring params returned by get_all_query_params_by_domain
+_qs_params = None
+
 # A LRUCache of domains to save us from having to do lots of regex matches
 _domain_cache = pylru.lrucache(500)
 
@@ -53,6 +65,7 @@ _domain_cache = pylru.lrucache(500)
 _naive_re = re.compile(r'\.?search\.')
 _naive_params = ('q', 'query', 'k', 'keyword', 'term',)
 
+
 def _unicode_parse_qs(qs, **kwargs):
     """
     A wrapper around ``urlparse.parse_qs`` that converts unicode strings to
@@ -60,11 +73,11 @@ def _unicode_parse_qs(qs, **kwargs):
     to latin-1 see http://hg.python.org/cpython/file/2.7/Lib/urlparse.py
 
     :param qs:       Percent-encoded query string to be parsed.
-    :type qs:        ``basestring``
+    :type qs:        ``str``
 
     :param kwargs:   Other keyword args passed onto ``parse_qs``.
     """
-    if isinstance(qs, str):
+    if PY3 or isinstance(qs, bytes):
         # Nothing to do
         return parse_qs(qs, **kwargs)
 
@@ -87,30 +100,30 @@ def _unicode_urlparse(url, encoding='utf-8', errors='ignore'):
     all elements of the parse result are unicode.
 
     :param url:      A URL.
-    :type url:       ``str``, ``unicode`` or :class:`urlparse.ParseResult`
+    :type url:       ``bytes``, ``unicode`` or :class:`urlparse.ParseResult`
 
     :param encoding: The string encoding assumed in the underlying ``str`` or
                      :class:`urlparse.ParseResult` (default is utf-8).
-    :type encoding:  ``str``
+    :type encoding:  ``bytes``
 
     :param errors:   response from ``decode`` if string cannot be converted to
                      unicode given encoding (default is ignore).
-    :type errors:    ``str``
+    :type errors:    ``bytes``
     """
-    if isinstance(url, str):
+    if isinstance(url, bytes):
         url = url.decode(encoding, errors)
     elif isinstance(url, ParseResult):
         # Ensure every part is unicode because we can't rely on clients to do so
         parts = list(url)
         for i in range(len(parts)):
-            if isinstance(parts[i], str):
+            if isinstance(parts[i], bytes):
                 parts[i] = parts[i].decode(encoding, errors)
         return ParseResult(*parts)
 
     try:
         return urlparse(url)
     except ValueError:
-        msg = u'Malformed URL "{}" could not parse'.format(url)
+        msg = 'Malformed URL "{}" could not parse'.format(url)
         log.debug(msg, exc_info=True)
         return None
 
@@ -126,7 +139,7 @@ def _serp_query_string(parse_result):
     """
     query = parse_result.query
     if parse_result.fragment != '':
-        query = u'{}&{}'.format(query, parse_result.fragment)
+        query = '{}&{}'.format(query, parse_result.fragment)
 
     return query
 
@@ -138,13 +151,13 @@ def _is_url_without_path_query_or_fragment(url_parts):
     :param url_parts: A URL.
     :type url_parts:  :class:`urlparse.ParseResult`
     """
-    return url_parts.path.strip('/') == '' and url_parts.query == '' \
+    return url_parts.path.strip('/') in ['', 'search'] and url_parts.query == '' \
            and url_parts.fragment == ''
 
 _engines = None
 def _get_search_engines():
     """
-    Convert the OrderedDict of search engine parsers that we get from Piwik
+    Convert the OrderedDict of search engine parsers that we get from Matomo
     to a dictionary of SearchEngineParser objects.
 
     Cache this thing by storing in the global ``_engines``.
@@ -153,97 +166,68 @@ def _get_search_engines():
     if _engines:
         return _engines
 
-    piwik_engines = _get_piwik_engines()
+    matomo_engines = _get_matomo_engines()
     # Engine names are the first param of each of the search engine arrays
     # so we group by those guys, and create our new dictionary with that
     # order
-    get_engine_name = lambda x: x[1][0]
-    definitions_by_engine = groupby(piwik_engines.iteritems(), get_engine_name)
     _engines = {}
 
-    for engine_name, rule_group in definitions_by_engine:
+    for engine_name, rule_group in iteritems(matomo_engines):
         defaults = {
             'extractor': None,
             'link_macro': None,
-            'charsets': ['utf-8']
+            'charsets': ['utf-8'],
+            'hiddenkeyword': None
         }
 
-        for i, rule in enumerate(rule_group):
-            domain = rule[0]
-            rule = rule[1][1:]
-            if i == 0:
-                defaults['extractor'] = rule[0]
-                if len(rule) >= 2:
-                    defaults['link_macro'] = rule[1]
-                if len(rule) >= 3:
-                    defaults['charsets'] = rule[2]
+        for rule in rule_group:
+            if any(url for url in rule['urls'] if '{}' in url):
+                rule['urls'] = _expand_country_codes(rule['urls'])
+            for i, domain in enumerate(rule['urls']):
+                if i == 0:
+                    defaults['extractor'] = rule['params']
+                    if 'backlink' in rule:
+                        defaults['link_macro'] = rule['backlink']
+                    if 'charsets' in rule:
+                        defaults['charsets'] = rule['charsets']
+                    if 'hiddenkeyword' in rule:
+                        defaults['hiddenkeyword'] = rule['hiddenkeyword']
 
                 _engines[domain] = SearchEngineParser(engine_name,
                                                       defaults['extractor'],
                                                       defaults['link_macro'],
-                                                      defaults['charsets'])
-                continue
-
-            # Default args for SearchEngineParser
-            args = [engine_name, defaults['extractor'],
-                    defaults['link_macro'], defaults['charsets']]
-            if len(rule) >= 1:
-                args[1] = rule[0]
-
-            if len(rule) >= 2:
-                args[2] = rule[1]
-
-            if len(rule) == 3:
-                args[3] = rule[2]
-
-            _engines[domain] = SearchEngineParser(*args)
+                                                      defaults['charsets'],
+                                                      defaults['hiddenkeyword'])
 
     return _engines
 
 
-def _get_piwik_engines():
+def _expand_country_codes(urls):
+    urls = set(urls) if isinstance(urls, list) else {urls}
+    expanded_urls = {url.format(country_code) for url in urls
+                     for country_code in _country_codes}
+    expanded_urls.update({url.format(second_level_domain + '.' + cc_sub_domain)
+                          for url in urls
+                          for cc_sub_domain in _country_codes
+                          for second_level_domain in _second_level_domains
+                          if not url[-1].isalnum()})
+    return expanded_urls
+
+
+def _get_matomo_engines():
     """
     Return the search engine parser definitions stored in this module. We don't
     cache this result since it's only supposed to be called once.
     """
     stream = pkg_resources.resource_stream
-    with stream(__name__, 'search_engines.pickle') as picklestream:
-        _piwik_engines = pickle.load(picklestream)
-
-    return _piwik_engines
-
-
-_get_lossy_domain_regex = None
-def _get_lossy_domain(domain):
-    """
-    A lossy version of a domain/host to use as lookup in the ``_engines``
-    dict.
-
-    :param domain: A string that is the ``netloc`` portion of a URL.
-    :type domain:  ``str``
-    """
-    global _domain_cache, _get_lossy_domain_regex
-
-    if domain in _domain_cache:
-        return _domain_cache[domain]
-
-    if not _get_lossy_domain_regex:
-        codes = '|'.join(_country_codes)
-        _get_lossy_domain_regex = re.compile(
-                r'^' # start of string
-                r'(?:w+\d*\.|search\.|m\.)*' + # www. www1. search. m.
-                r'((?P<ccsub>{})\.)?'.format(codes) + # country-code subdomain
-                r'(?P<domain>.*?)' + # domain
-                r'(?P<tld>\.(com|org|net|co|it|edu))?' + # tld
-                r'(?P<tldcc>\.({}))?'.format(codes) + # country-code tld
-                r'$') # all done
-
-    res = _get_lossy_domain_regex.match(domain).groupdict()
-    output = u'%s%s%s' % ('{}.' if res['ccsub'] else '',
-                          res['domain'],
-                          '.{}' if res['tldcc'] else res['tld'] or '')
-    _domain_cache[domain] = output # Add to LRU cache
-    return output
+    with stream(__name__, 'search_engines.json') as json_stream:
+        if PY3:
+            if hasattr(json_stream, 'buffer'):
+                json_stream = TextIOWrapper(json_stream.buffer, encoding='utf-8')
+            else:
+                json_stream = TextIOWrapper(json_stream, encoding='utf-8')
+        _matomo_engines = json.load(json_stream)
+    return _matomo_engines
 
 
 class ExtractResult(object):
@@ -260,20 +244,22 @@ class ExtractResult(object):
 
 
 class SearchEngineParser(object):
-    """Handles persing logic for a single line in Piwik's list of search
+    """Handles persing logic for a single line in Matomo's list of search
     engines.
 
-    Piwik's list for reference:
+    Matomo's list for reference:
 
-    https://raw.github.com/piwik/piwik/master/core/DataFiles/SearchEngines.php
+    https://raw.githubusercontent.com/matomo-org/searchengine-and-social-list/master/SearchEngines.yml
 
     This class is not used directly since it already assumes you know the
     exact search engine you want to use to parse a URL. The main interface
     for users of this module is the :func:`extract` method.
     """
-    __slots__ = ('engine_name', 'keyword_extractor', 'link_macro', 'charsets')
+    __slots__ = ('engine_name', 'keyword_extractor', 'link_macro', 'charsets',
+                 'hidden_keyword_paths')
 
-    def __init__(self, engine_name, keyword_extractor, link_macro, charsets):
+    def __init__(self, engine_name, keyword_extractor, link_macro, charsets,
+                 hidden_keyword_paths=None):
         """New instance of a :class:`SearchEngineParser`.
 
         :param engine_name:         the friendly name of the engine (e.g.
@@ -292,9 +278,16 @@ class SearchEngineParser(object):
 
         :param charsets:            a string or list of charsets to use to
                                     decode the keyword
+
+        :param hidden_keywords_paths: an optional list of strings (that may
+                                      contain regular expressions) describing
+                                      valid paths for the search engine that may
+                                      not contain any keywords. Regular
+                                      expressions are expected to be surround by
+                                      `/` characters.
         """
         self.engine_name = engine_name
-        if isinstance(keyword_extractor, basestring):
+        if isinstance(keyword_extractor, string_types):
             keyword_extractor = [keyword_extractor]
         self.keyword_extractor = keyword_extractor[:]
         for i, extractor in enumerate(self.keyword_extractor):
@@ -305,9 +298,19 @@ class SearchEngineParser(object):
                 self.keyword_extractor[i] = extractor
 
         self.link_macro = link_macro
-        if isinstance(charsets, basestring):
+        if isinstance(charsets, string_types):
             charsets = [charsets]
         self.charsets = [c.lower() for c in charsets]
+        if hidden_keyword_paths:
+            self.hidden_keyword_paths = hidden_keyword_paths[:]
+        else:
+            self.hidden_keyword_paths = []
+        for i, path in enumerate(self.hidden_keyword_paths):
+            # Pre-compile all the regular expressions
+            if len(path) > 1 and path.startswith('/') and path.endswith('/'):
+                path = path[1:-1]
+                path = re.compile(path)
+                self.hidden_keyword_paths[i] = path
 
     def get_serp_url(self, base_url, keyword):
         """
@@ -324,7 +327,7 @@ class SearchEngineParser(object):
         if self.link_macro is None:
             return None
 
-        link = u'{}/{}'.format(base_url, self.link_macro.format(k=keyword))
+        link = '{}/{}'.format(base_url, self.link_macro.format(k=keyword))
         return link
 
     def parse(self, url_parts):
@@ -351,8 +354,7 @@ class SearchEngineParser(object):
             # e.g. &prev=/search%3Fq%3Dimages%26sa%3DX%26biw%3D320%26bih%3D416%26tbm%3Disch
             engine_name = 'Google Images'
             if 'prev' in query:
-                prev_query = _unicode_parse_qs(urlparse(query['prev'][0]).query)
-                keyword = prev_query.get('q', [None])[0]
+                query = _unicode_parse_qs(_unicode_urlparse(query['prev'][0]).query)
         elif engine_name == 'Google' and 'as_' in original_query:
             # Google has many different ways to filter results.  When some of
             # these filters are applied, we can no longer just look for the q
@@ -364,25 +366,25 @@ class SearchEngineParser(object):
             # Search Operator: None (same as normal search)
             key = query.get('as_q')
             if key:
-              keys.append(key[0])
+                keys.append(key[0])
             # Results should contain any of these words
             # Search Operator: <keyword> [OR <keyword>]+
             key = query.get('as_oq')
             if key:
-              key = key[0].replace('+', ' OR ')
-              keys.append(key)
+                key = key[0].replace('+', ' OR ')
+                keys.append(key)
             # Results should match the exact phrase
             # Search Operator: "<keyword>"
             key = query.get('as_epq')
             if key:
-              keys.append(u'"{}"'.format(key[0]))
+                keys.append('"{}"'.format(key[0]))
             # Results should contain none of these words
             # Search Operator: -<keyword>
             key = query.get('as_eq')
             if key:
-              keys.append(u'-{}'.format(key[0]))
+                keys.append('-{}'.format(key[0]))
 
-            keyword = u' '.join(keys).strip()
+            keyword = ' '.join(keys).strip()
 
         if engine_name == 'Google':
             # Check for usage of Google's top bar menu
@@ -400,7 +402,7 @@ class SearchEngineParser(object):
 
         # Otherwise we keep looking through the defined extractors
         for extractor in self.keyword_extractor:
-            if not isinstance(extractor, basestring):
+            if not isinstance(extractor, string_types):
                 # Regular expression extractor
                 match = extractor.search(url_parts.path)
                 if match:
@@ -413,28 +415,45 @@ class SearchEngineParser(object):
                     # most recent
                     keyword = query[extractor][-1]
 
-                # Now we have to check for a tricky case where it is a SERP
-                # but just with no keyword as can be the case with Google
-                # Images or DuckDuckGo
-                if keyword is None and extractor == 'q' and \
-                   engine_name in ('Google Images', 'DuckDuckGo'):
-                    keyword = ''
-                elif keyword is None and extractor == 'q' and \
-                     engine_name == 'Google' and \
-                     _is_url_without_path_query_or_fragment(url_parts):
-                    keyword = ''
+                # Now we have to check for a tricky case where it is a SERP but
+                # there are no keywords
+                if keyword == '':
+                    keyword = False
+
+                if keyword is not None:
+                    break
+
+        # if no keyword found, but empty/hidden keywords are allowed
+        if self.hidden_keyword_paths and (keyword is None or keyword is False):
+            path_with_query_and_frag = url_parts.path
+            if url_parts.query:
+                path_with_query_and_frag += '?{}'.format(url_parts.query)
+            if url_parts.fragment:
+                path_with_query_and_frag += '#{}'.format(url_parts.fragment)
+            for path in self.hidden_keyword_paths:
+                if not isinstance(path, string_types):
+                    if path.search(path_with_query_and_frag):
+                        keyword = False
+                        break
+                elif path == path_with_query_and_frag:
+                    keyword = False
+                    break
 
         if keyword is not None:
+            # Replace special placeholder with blank string
+            if keyword is False:
+                keyword = ''
             return ExtractResult(engine_name, keyword, self)
 
     def __repr__(self):
         repr_fmt = ("SearchEngineParser(engine_name={!r}, "
-                    "keyword_extractor={!r}, link_macro={!r}, charsets={!r})")
-        return repr_fmt.format(
-                        self.engine_name,
-                        self.keyword_extractor,
-                        self.link_macro,
-                        self.charsets)
+                    "keyword_extractor={!r}, link_macro={!r}, charsets={!r}, "
+                    "hidden_keywords={!r})")
+        return repr_fmt.format(self.engine_name,
+                               self.keyword_extractor,
+                               self.link_macro,
+                               self.charsets,
+                               self.hidden_keyword_paths)
 
 
 def add_custom_parser(match_rule, parser):
@@ -448,7 +467,7 @@ def add_custom_parser(match_rule, parser):
     :param parser:     A custom parser.
     :type parser:      :class:`SearchEngineParser`
     """
-    assert isinstance(match_rule, unicode)
+    assert isinstance(match_rule, text_type)
     assert isinstance(parser, SearchEngineParser)
 
     global _engines
@@ -466,15 +485,38 @@ def get_all_query_params():
     """
     engines = _get_search_engines()
     all_params = set()
-    _not_regex = lambda x: isinstance(x, basestring)
-    for parser in engines.itervalues():
+    for parser in itervalues(engines):
         # Find non-regex params
-        params = set(filter(_not_regex, parser.keyword_extractor))
+        params = {param for param in parser.keyword_extractor
+                  if isinstance(param, string_types)}
         all_params |= params
 
     return list(all_params)
 
 
+def get_all_query_params_by_domain():
+    """
+    Return all the possible query string params for all search engines.
+
+    :returns: a ``list`` of all the unique query string parameters that are
+              used across the search engine definitions.
+    """
+    global _qs_params
+    if _qs_params:
+        return _qs_params
+    engines = _get_search_engines()
+    param_dict = defaultdict(list)
+    for domain, parser in iteritems(engines):
+        # Find non-regex params
+        params = {param for param in parser.keyword_extractor
+                  if isinstance(param, string_types)}
+        tld_res = tldextract.extract(domain)
+        domain = tld_res.registered_domain
+        param_dict[domain] = sorted(set(param_dict[domain]) | params)
+    _qs_params = param_dict
+    return param_dict
+
+
 def get_parser(referring_url):
     """
     Utility function to find a parser for a referring URL if it is a SERP.
@@ -494,24 +536,19 @@ def get_parser(referring_url):
 
     domain = url_parts.netloc
     path = url_parts.path
-    lossy_domain = _get_lossy_domain(url_parts.netloc)
     engine_key = url_parts.netloc
-
+    stripped_domain = domain[4:] if domain.startswith('www.') else None
     # Try to find a parser in the engines list.  We go from most specific to
     # least specific order:
     # 1. <domain><path>
-    # 2. <lossy_domain><path>
-    # 3. <lossy_domain>
-    # 4. <domain>
-    # The final case has some special exceptions for things like Google custom
+    # 2. <custom search engines>
+    # 3. <domain>
+    # 4. <stripped_domain>
+    # The second step has some special exceptions for things like Google custom
     # search engines, yahoo and yahoo images
-    if u'{}{}'.format(domain, path) in engines:
-        engine_key = u'{}{}'.format(domain, path)
-    elif u'{}{}'.format(lossy_domain, path) in engines:
-        engine_key = u'{}{}'.format(lossy_domain, path)
-    elif lossy_domain in engines:
-        engine_key = lossy_domain
-    elif domain not in engines:
+    if '{}{}'.format(domain, path) in engines:
+        engine_key = '{}{}'.format(domain, path)
+    elif domain not in engines and stripped_domain not in engines:
         if query[:14] == 'cx=partner-pub':
             # Google custom search engine
             engine_key = 'google.com/cse'
@@ -527,7 +564,7 @@ def get_parser(referring_url):
         else:
             return None
 
-    return engines.get(engine_key)
+    return engines.get(engine_key) or engines.get(stripped_domain)
 
 
 def is_serp(referring_url, parser=None, use_naive_method=False):
@@ -606,7 +643,6 @@ def extract(serp_url, parser=None, lower_case=True, trimmed=True,
             query = _unicode_parse_qs(url_parts.query, keep_blank_values=True)
             for param in _naive_params:
                 if param in query:
-                    import tldextract
                     tld_res = tldextract.extract(url_parts.netloc)
                     return ExtractResult(tld_res.domain,
                                          query[param][0],
@@ -631,13 +667,11 @@ def extract(serp_url, parser=None, lower_case=True, trimmed=True,
 
 def main():
     import argparse
-    import sys
-    import re
 
     parser = argparse.ArgumentParser(
         description='Parse a SERP URL to extract engine name and keyword.')
 
-    parser.add_argument('input', metavar='url', type=unicode, nargs='*',
+    parser.add_argument('input', metavar='url', type=text_type, nargs='*',
                         help='A potential SERP URL')
     parser.add_argument('-l', '--list', default=False, action='store_true',
                         help='Print a list of all the SearchEngineParsers.')
@@ -646,11 +680,11 @@ def main():
 
     if args.list:
         engines = _get_search_engines()
-        engines = sorted(engines.iteritems(), key=lambda x: x[1].engine_name)
-        print '{:<30}{}'.format('Fuzzy Domain', 'Parser')
+        engines = sorted(iteritems(engines), key=lambda x: x[1].engine_name)
+        print('{:<30}{}'.format('Fuzzy Domain', 'Parser'))
         for fuzzy_domain, parser in engines:
-            print '{:<30}{}'.format(fuzzy_domain, parser)
-        print '{} parsers.'.format(len(engines))
+            print('{:<30}{}'.format(fuzzy_domain, parser))
+        print('{} parsers.'.format(len(engines)))
         sys.exit(0)
 
     if len(args.input) == 0:
@@ -665,8 +699,8 @@ def main():
             res = ['""', '""']
         else:
             res = [escape_quotes(res.engine_name), escape_quotes(res.keyword)]
-            res = [u'"{}"'.format(r) for r in res]
-        print u','.join(res)
+            res = ['"{}"'.format(r) for r in res]
+        print(','.join(res))
 
 if __name__ == '__main__':
     main()
